import numpy as np
from numpy.linalg import norm, eigh, qr
from scipy.linalg import schur


##### Matrix operations #########################################################
def Dag(U):
    return np.transpose(np.conj(U))

def vecxvec( E, wf, alpha): # Many times faster than multiply
    s = E.size
    for i in range(s):
        wf[i] = np.exp(-1j * E[i] * alpha) * wf[i]
    return wf
# Construct matrix exponentials and products faster than in numpy or scipy
def vm_exp_mul(E, V, dt = 1.0): # expm(diag(vector))*matrix  multiplication via exp(vector)*matrix
    s = E.size
    A = np.empty((s, s), np.complex128)
    for i in range(s):
        A[i,:] = np.exp(-1j * E[i] * dt) * Dag(V[:,i])
    return A
def expmH_from_Eig(E, V, dt = 1.0):
    U = np.dot(V, vm_exp_mul(E, V, dt))
    return U
def expmH(H, dt = 1.0):
    E, V = eigh(H)
    return expmH_from_Eig( E, V, dt)

##### Logarithms 
def unitary_eig(A): # alternative to np.eig returning unitary matrices V
    Emat, V = schur(A, output='complex')
    return np.diag(Emat), V
def vm_log_mul(E, V):
    s = E.size
    A = np.empty((s, s), np.complex128)
    for i in range(s):
        A[i,:] = np.log(E[i])*Dag(V[:,i])
    return A
def logmU(U):
    E, V = unitary_eig(U)
    return np.dot(V, vm_log_mul(E, V))

def logmU_parts(U):
    E, V = unitary_eig(U)
    return -np.log(E).imag, V

##### Scripts to create test interpolation #######################################
def H_from_c(H_s, c):
    H = H_s[0].copy()
    for i in range(len(c)):
        H = H + c[i]*H_s[i+1]
    return H

# Get unitary from weights c and H_s
def expm_H_s(H_s, c):
    H = H_from_c(H_s, c)
    E, V = eigh(H)
    return expmH_from_Eig( E, V )

# Construct the terms for corners of an interpolation (hyper-) cube
def make_border_unitaries(H_s, c_mins, c_maxs):
    if isinstance(c_mins, (list)):
        c_mins = np.array(c_mins, dtype=np.float64)
    if isinstance(c_maxs, (list)):
        c_maxs = np.array(c_maxs, dtype=np.float64)
    n = len(c_mins)
    s = H_s.shape[-1]
    U0 = expm_H_s(H_s, c_mins)
    Ui = np.empty((n, s, s), dtype=np.complex128)
    for i in range(n):
        curr_c = c_mins.copy()
        curr_c[i] = c_maxs[i]
        Ui[i] = expm_H_s(H_s, curr_c)
    n_ij = n * (n-1) // 2
    Uij_2 = np.empty((n_ij, s, s), dtype=np.complex128)
    Ui_2 = np.empty((n, s, s), dtype=np.complex128)
    ind = 0
    dc = (c_maxs - c_mins) 
    for i in range(n):
        curr_c = c_mins.copy()
        curr_c[i] += dc[i]/2
        Ui_2[i] = expm_H_s(H_s, curr_c)
        for j in range(i+1, n):
            curr_c = c_mins.copy()
            curr_c[i] += dc[i]/2
            curr_c[j] += dc[j]/2
            Uij_2[ind] = expm_H_s(H_s, curr_c)
            ind += 1
    curr_c = c_mins + dc/2
    U_2 = expm_H_s(H_s, curr_c)
    return U0, Ui, Ui_2, Uij_2, U_2

# Get the unitary interpolation along 
# Construct interpolation terms from U0 to Ui
def make_interpolation_terms(U0, Ui):
    n = Ui.shape[0]
    s = Ui.shape[-1]
    Es = np.empty((n,s), dtype=np.float64)
    Vs = np.empty((n,s,s), dtype=np.complex128)
    for i in range(n):
        U = np.dot(Ui[i], Dag(U0))
        E, V = logmU_parts(U)
        Es[i] = E
        Vs[i] = V
    return Es, Vs

def interpolation_core_term(Vs, Es, index, alpha):
    E = Es[index]
    V = Vs[index]
    return V @ np.diag(np.exp(-1j*E*alpha)) @ Dag(V)

def nd_interpolation_core_term(Vs, Es, U0, alphas, indices):
    U = interpolation_core_term(Vs, Es, indices[0], alphas[0])
    for i in range(1, len(indices)):
        U = np.dot(interpolation_core_term(Vs, Es, indices[i], alphas[i]), U)
    return U @ U0

###### Scripts for caching values for polynomial approximation of Infidelities from Test binning ######################
# test directional accuracy of interpolation
def UI_directional_overlap_traces(U0, Ui_2, Uij_2, U_2, Es, Vs):
    # Test the quality of the interpolation for every combination (i,j) of directions at point x_i = 0.5 = x_j
    # exception 1d: test only center
    # Calculate and return the trace of the overlap operator M = U_exact^dagger @ U_interpolated (-d) , where d is the hilbert space dimension, which is subtracted except for all tr_M except the central element tr_M_2
    n = Es.shape[0]
    d = U0.shape[0] 
    if n == 1:
        alphas = np.array([0.5])
        indices = np.array([0])
        U = nd_interpolation_core_term(Vs, Es, U0, alphas, indices)
        tr_M = np.array([])
        tr_M_single = np.array([np.trace(Dag(Ui_2[0]) @ U) - d])
        tr_M_2 = tr_M_single[0]
        indexes = np.empty((0,0), dtype=np.compat.long)
    else:
        n_ij = n * (n-1) // 2
        tr_M = np.empty(n_ij, dtype=np.complex128)
        tr_M_single = np.empty(n, dtype=np.complex128)
        indexes = np.empty((n_ij, 2), dtype=np.compat.long)
        ind = 0
        alphas = np.array([0.5])
        for i in range(n):
            U = nd_interpolation_core_term(Vs, Es, U0, alphas, np.array([i]))
            tr_M_single[i] = np.trace(Dag(Ui_2[i]) @ U) - d
        alphas = np.array([0.5, 0.5])
        for i in range(n):
            for j in range(i+1, n):
                indices = np.array([i,j])
                U = nd_interpolation_core_term(Vs, Es, U0, alphas, indices)
                tr_M[ind] = np.trace(Dag(Uij_2[ind]) @ U) - d - tr_M_single[i] - tr_M_single[j]
                indexes[ind] = np.array([i,j], dtype=np.compat.long)
                ind += 1
    tr_M = tr_M + tr_M.conj()
    tr_M = tr_M.real.astype(np.float64)
    tr_M_single = tr_M_single + tr_M_single.conj()
    tr_M_single = tr_M_single.real.astype(np.float64)
    # compare central element
    if n > 2:
        alphas = np.array([0.5]*n)
        indices = np.arange(n)
        U = nd_interpolation_core_term(Vs, Es, U0, alphas, indices)
        tr_M_2 = np.trace(Dag(U_2) @ U) - d
        tr_M_2 = tr_M_2 + tr_M_2.conj()
        tr_M_2 = tr_M_2.real.astype(np.float64)
    elif n == 2:
        tr_M_2 = tr_M[0]
    else:
        tr_M_2 = tr_M_single[0]
    pref = -1/(d+1)
    tr_M = pref * tr_M
    tr_M_single = pref * tr_M_single
    tr_M_2 = pref * tr_M_2
    return tr_M_single, tr_M, tr_M_2, indexes

def get_bidirectional_overlap_operators(H_s, c_mins=None, c_maxs=None, bins=None): # Calculate the infidelity parameters for a given interpolation
    n = H_s.shape[0]-1
    if c_mins is None:
        c_mins = np.zeros(n)
    if c_maxs is None:
        c_maxs = np.ones(n)
    if bins is None:
        bins = np.ones(n, dtype=np.compat.long)
    else:
        bins = bins.astype(np.compat.long)
    dc = (c_maxs - c_mins)
    c_maxs = c_mins + dc / bins
    U0, Ui, Ui_2, Uij_2, U_2_exact = make_border_unitaries(H_s, c_mins, c_maxs)
    Es, Vs = make_interpolation_terms(U0, Ui)
    tr_M_single, tr_M, tr_M_2, indexes = UI_directional_overlap_traces(U0, Ui_2, Uij_2, U_2_exact, Es, Vs)  # , tr_M_2
    alphas = np.array([0.5]*n)
    indices = np.arange(n)
    U_2_approx = nd_interpolation_core_term(Vs, Es, U0, alphas, indices)
    return tr_M_single, tr_M, tr_M_2, indexes, U_2_exact, U_2_approx

def I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, param=None, bins=None):
    if bins is None:
        bins = np.ones(n, dtype=np.compat.long)
    if param is None:
        param = np.ones(n)
    param = param / bins
    res = 0.0
    if len(indexes) > 0:
        multipliers_2 = param[indexes[:,0]]**2*param[indexes[:,1]]**2
        res += np.sum(multipliers_2 * tr_M)
    multipliers_4 = param**4
    res +=  np.sum(multipliers_4 * tr_M_single)
    return res

def I_from_tr_M(tr_M_single, tr_M, indexes, n, param=None):
    return I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, param=param, bins=None)

##### Scripts for Optimization ########################################################
def cache_size_by_bins(bins):
    # calculate the number of cache elements for a given binning
    n = bins.shape[0] #len(bins)
    if n > 1:
        n_C = 2**(n-1) * np.prod(bins)
        n_L = bins[0] * np.prod(bins[1:] + 1)
        n_R = bins[-1] * np.prod(bins[:-1] + 1) 
        N_tot = n_C + n_L + n_R
    else:
        N_tot = bins[0]
    return N_tot

def cache_size_change_by_dir(bins, negative=False):
    # check by how much the cache grows for every direction it could change in
    n = bins.shape[0] #len(bins)
    curr_size = cache_size_by_bins(bins)
    differences = np.zeros(len(bins))
    for i in range(n):
        curr_bins = bins.copy()
        if not negative:
            curr_bins[i] += 1
        else:
            curr_bins[i] -= 1
        new_size = cache_size_by_bins(curr_bins)
        differences[i] = new_size - curr_size
    return differences

def infidelity_change_by_dir(tr_M_single, tr_M, indexes, n, bins, negative=False):
    # check by how much the infidelity changes for every direction it could change in
    curr_I = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bins=bins)
    differences = np.zeros(len(bins))
    for i in range(n):
        curr_bins = bins.copy()
        if not negative:
            curr_bins[i] += 1
        else:
            curr_bins[i] -= 1
        new_I = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bins=curr_bins)
        differences[i] = curr_I - new_I 
    return differences

def _find_solution(tr_M_single, tr_M, indexes, n, bins, I_tar=10**-14):
    # increase to find solution
    curr_I = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bins=bins)
    while curr_I > I_tar:
        inf_diff = infidelity_change_by_dir(tr_M_single, tr_M, indexes, n, bins)
        cache_differences = cache_size_change_by_dir(bins)
        curr_prices_of_optimization = inf_diff / cache_differences # infidelity decrease per cache increase
        # any of the new points smaller than the target?
        new_Is = curr_I - inf_diff
        if np.any(new_Is < I_tar):
            # find the one with the smallest increase in cache size
            inds = np.where(new_Is < I_tar)
            min_index = np.argmin(cache_differences[inds])
            min_dir = inds[0][min_index]
            bins[min_dir] += 1
            curr_I = new_Is[min_dir]
        else:
            # find the one with the smallest cache price increase for the largest infidelity decrease
            max_ind = np.argmax(curr_prices_of_optimization)
            bins[max_ind] += 1
            curr_I = new_Is[max_ind]
    return bins, curr_I
def _optimize_solution(tr_M_single, tr_M, indexes, n, bins, I_tar=10**-14):
    # decrease for better solution
    optimizing = True # while the current binnning is changing keep optimizing= True
    curr_I = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bins=bins)
    curr_cache = cache_size_by_bins(bins)
    if curr_I > I_tar:
        raise Exception('Initial solution not < I_tar')
    new_Is = np.zeros(n)
    new_caches = np.zeros(n, dtype=np.compat.long)
    while optimizing:
        # check neighbors I
        for i in range(n):
            curr_bins = bins.copy()
            curr_bins[i] -= 1
            if curr_bins[i] < 1:
                new_Is[i] = 1.0
            else:
                new_Is[i] = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bins=curr_bins)
            # cache smaller?
            new_caches[i] = cache_size_by_bins(curr_bins)
        # find I < I_tar among now_Is
        inds = np.where(new_Is < I_tar)[0]
        if len(inds) > 0:
            min_index = np.argmin(new_caches[inds])
            potential_ind = inds[min_index]
            if new_caches[potential_ind] < curr_cache:
                bins[potential_ind] -= 1
                curr_I = new_Is[potential_ind]
                curr_cache = new_caches[potential_ind]
            else:
                optimizing = False
        else:
            optimizing = False
    return bins, curr_I

def separate_indexes(i, indexes, tr_M):
    # get indexes that contain i
    inds_i = np.where(indexes[:,0] == i)[0]
    inds_i = np.concatenate((inds_i, np.where(indexes[:,1] == i)[0]))
    # get indexes that dont contain i, the inds not in inds_i
    inds_not_i = np.setdiff1d(np.arange(len(indexes)), inds_i)
    indexes_not_i = indexes[inds_not_i]
    ind_i = indexes[inds_i]
    # reduce indexes_i to not contain i, make it a one d array
    indexes_i = np.empty(len(ind_i), dtype=np.compat.long)
    for j in range(len(ind_i)):
        indexes_i[j] = ind_i[j][0] if ind_i[j][0] != i else ind_i[j][1]
    tr_M_not_i = tr_M[inds_not_i]
    tr_M_i = tr_M[inds_i]
    return indexes_i, indexes_not_i, tr_M_i, tr_M_not_i

def separate_all_indexes(indexes, tr_M, n):	
    indexes_i = []
    indexes_not_i = []
    tr_M_i = []
    tr_M_not_i = []
    for i in range(n):
        inds_i, inds_not_i, tr_M_i_, tr_M_not_i_ = separate_indexes(i, indexes, tr_M)
        indexes_i.append(inds_i)
        indexes_not_i.append(inds_not_i)
        tr_M_i.append(tr_M_i_)
        tr_M_not_i.append(tr_M_not_i_)
    return indexes_i, indexes_not_i, tr_M_i, tr_M_not_i

def I_rest_from_tr_M_bins(i, tr_M_single, tr_M_not_i, indexes_not_i, n, bins):
    if bins is None:
        bins = np.ones(n, dtype=np.compat.long)
    param = 1 / bins
    res = 0.0
    if len(indexes_not_i) > 0:
        multipliers_2 = param[indexes_not_i[:,0]]**2*param[indexes_not_i[:,1]]**2
        res += np.sum(multipliers_2 * tr_M_not_i)
    multipliers_4 = param**4
    res +=  np.sum(multipliers_4[:i] * tr_M_single[:i]) + np.sum(multipliers_4[i+1:] * tr_M_single[i+1:])
    return res

def I_i_from_tr_M_bins(i, tr_M_single, tr_M_i, indexes_i, n, bins):
    if bins is None:
        bins = np.ones(n, dtype=np.compat.long)
    param = 1 / bins
    res = 0.0
    if len(indexes_i) > 0:
        multipliers_2 = param[indexes_i]**2
        res += np.sum(multipliers_2 * tr_M_i) * param[i]**2
    res +=  param[i]**4 * tr_M_single[i]
    return res

# optimize solution by reducing one bin direction by one and then finding the optimum along the other directions 
def optimum_along_single_direction(i, tr_M_single, tr_M_i, tr_M_not_i, indexes_i, indexes_not_i, n, bins, I_tar=10**-14):
    # rewrite as qudratic expression, to get optimum along one direction
    new_bins = bins.copy()
    success = False
    s = 1/bins**2
    Pii =tr_M_single[i]
    I = I_rest_from_tr_M_bins(i, tr_M_single, tr_M_not_i, indexes_not_i, n, bins) - I_tar
    if I < 0:
        Ci = np.sum(tr_M_i * s[indexes_i])
        # find optimum along s_i -> quadratic expression in s_i
        Ci_Pii = Ci/Pii/2
        sqrt_internal = Ci_Pii**2 - I/Pii
        if sqrt_internal > 0:
            sqrt_term = np.sqrt(sqrt_internal)
            s_i = - Ci_Pii - sqrt_term
            if s_i < 0:
                s_i = - Ci_Pii + sqrt_term
            # find the 
            success = True if s_i > 0 else False
            if success:
                # from s_i to_bins[i]
                bins_i = np.sqrt(1/s_i)
                new_bins = bins.copy()
                new_bins[i] = np.ceil(bins_i).astype(np.compat.long)
    return new_bins, success

##### The script we need to run to optimize binning
def optimize_binning(tr_M_single, tr_M, indexes, n, bins=None, I_tar=10**-14):
    # optimize binning to get minimum cache for target infidelity
    # while I > I_tar: add bins in a direction that increases infidelity the most per cache size, if more than one direction would achieve the target fidelity 
    if bins is None:
        bins = np.ones(n, dtype=np.compat.long)
    bins, curr_I = _find_solution(tr_M_single, tr_M, indexes, n, bins, I_tar)
    # check surrounding points for smaller cache size
    bins, curr_I = _optimize_solution(tr_M_single, tr_M, indexes, n, bins, I_tar)

    # separate indexes and tr_M
    if len(indexes) > 0:
        indexes_i, indexes_not_i, tr_M_i, tr_M_not_i = separate_all_indexes(indexes, tr_M, n)
        # for every index i, reduce by one, check if solution is still in target range,
        # if not, find which directions first inner point has lower cache size. if non has lower cache size, stop and go back to previous solution
        last_bins = bins.copy()
        found_lower = True
        curr_min_cache = cache_size_by_bins(bins)
        while found_lower:
            found_lower = False
            for i in range(n):
                bin_red = last_bins.copy()
                if bin_red[i] > 1:
                    bin_red[i] -= 1
                    #I_bin_red = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bin_red)
                    #cache_bin_red = cache_size_from_bins(bin_red)
                    # check for every other direction, which point leads to I < I_tar, and save the cache needed
                    for j in range(n):
                        if j == i:
                            continue
                        else:
                            bin_j, success = optimum_along_single_direction(j, tr_M_single, tr_M_i[j], tr_M_not_i[j], indexes_i[j], indexes_not_i[j], n, bin_red, I_tar)
                            if success:
                                I_bin_j = I_from_tr_M_bins(tr_M_single, tr_M, indexes, n, bins=bin_j)
                                if not I_bin_j < I_tar:
                                    raise ValueError('I_bin_j='+str(I_bin_j)+' should be smaller than I_tar, algorithm is broken. (' + str(bin_red) + '->' + str(bin_j) + ')')
                                cache_bin_j = cache_size_by_bins(bin_j)
                                if cache_bin_j < curr_min_cache:
                                    curr_min_cache = cache_bin_j
                                    bins = bin_j
                                    found_lower = True
            if found_lower:
                last_bins = bins.copy()
        bins = last_bins
    return bins, curr_I


def optimal_binning(H_s, c_mins, c_maxs, I_tar):
    n = H_s.shape[0] 
    n -= 1
    bins = np.ones(n, dtype=np.compat.long)
    tr_M_single, tr_M, tr_M_2, indexes, U_exact, U_approx = get_bidirectional_overlap_operators(H_s, c_mins, c_maxs, bins=bins)
    opt_bins, curr_I = optimize_binning(tr_M_single, tr_M, indexes, n, bins=bins, I_tar=I_tar)
    return opt_bins


cdef void MM_cdot_pointer_v_scaled(double complex *a0, double complex *v0, double complex *c0, double complex alpha, int n, int m) nogil
cdef void MM_cdot_pointer_v(double complex *a0, double complex *v0, double complex *c0, int n, int m) nogil
cpdef void MM_cdot_v(double complex[:,::1] A, double complex[:,::1] v, double complex[:,::1] C)
cdef void DagM_M_cdot_pointer_v(double complex *a0, double complex *v0, double complex *c0, int n, int m) nogil
cpdef void DagM_M_cdot_v(double complex[:,::1] A, double complex[:,::1] v, double complex[:,::1] C) nogil

cdef void v_exp_pointer_v(double *e0, double complex *v0, double t, int n, int m) nogil
cdef void v_exp_v_pointer_v(double amp, double *e0, double complex *v0, double t, int n, int m) nogil

cdef void MM_cdot_pointer_batch_v(double complex *a0, double complex *vi, double complex *vo, double complex *di, double complex *do, int n, int m, int nm, int batch_ind) nogil
cdef void MM_cdot_batch_v(double complex[:,::1] A, double complex[:,::1] V_in, double complex[:,::1] V_out, double complex[:,:,::1] dV_in, double complex[:,:,::1] dV_out, int n, int m, int nm, int batch_ind) nogil

cdef void v_exp_v_and_v_exp_pointer_v_batch(double *e0, double complex *e0s, double complex *v0, double complex *dv0, double t, int n, int m, int batch_ind) nogil
cpdef void v_exp_v_and_v_exp_v_batch(double[::1] E, double complex[::1] Es, double complex[:,::1] V, double complex[:,:,::1] dV, double t, int batch_ind)
cdef void v_exp_v_and_v_exp_pointer_v_batch_increment(double amp, double *e0, double complex *e0s, double complex *v0, double complex *dv0, double t, int n, int m, int mn, int batch_ind) nogil
cpdef void v_exp_v_and_v_exp_v_batch_increment(double amp, double[::1] E, double complex[::1] Es, double complex[:,::1] V, double complex[:,:,::1] dV, double t, int batch_ind)

cdef double norm_inf_complex( double complex[:, ::1] B, int d, int m) nogil
cpdef expm_multiply_prepare(double complex[:,::1] A0, int m, double tol)


"""Compute the action of the matrix exponential onto a vector of a few columns of vectors."""
# Cythonized the _expm_multiply_simple_core function from scipy.sparse.linalg._expm_multiply
import numpy as np
from libc.math cimport exp
from scipy.linalg.cython_blas cimport zgemm, zaxpy
from unipolator.blas_functions cimport c_mat_scale
from unipolator.exp_and_log cimport copy_pointer 
from scipy.sparse.linalg._expm_multiply import LazyOperatorNormInfo, _fragment_3_1

# a function that prepares a expm_multiply_simple_core call by determining the number of iterations and matrix multiplications
cdef expm_multiply_prepare(A, m, t=1.0):
    """
    Prepare the computation of the action of the matrix exponential of A on B.

    Parameters
    ----------
    A : transposable linear operator
        The operator whose exponential is of interest.
    m : number of wavevectors to use (B will be n x m)
    t : float (default=1.0)
    References
    ----------
    .. [1] Awad H. Al-Mohy and Nicholas J. Higham (2011)
           "Computing the Action of the Matrix Exponential,
           with an Application to Exponential Integrators."
           SIAM Journal on Scientific Computing,
           33 (2). pp. 488-511. ISSN 1064-8275
           http://eprints.ma.man.ac.uk/1591/

    .. [2] Nicholas J. Higham and Awad H. Al-Mohy (2010)
           "Computing Matrix Functions."
           Acta Numerica,
           19. 159-208. ISSN 0962-4929
           http://eprints.ma.man.ac.uk/1451/
    """
    # Check the inputs.
    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError('expected A to be like a square matrix')
    ident = np.eye(A.shape[0], A.shape[1], dtype=A.dtype)
    n = A.shape[0]
    traceA =  A.trace()
    mu = traceA / float(n)
    A = A - mu * ident
    A_1_norm = np.linalg.norm(A, 1)  
    if t*A_1_norm == 0:
        m_star, s = 0, 1
    else:
        norm_info = LazyOperatorNormInfo(t*A, A_1_norm=t*A_1_norm)
        m_star, s = _fragment_3_1(norm_info, m)
    return m_star, s, mu

cdef void c_mat_add_pointer(double complex *a0, double complex *b0, int nn) nogil: # A = A+B*c
    cdef int incz = 1
    cdef double complex c = 1.0 + 0.0j
    zaxpy(&nn, &c, b0, &incz, a0, &incz)

cpdef double norm_inf_complex(double complex[:, ::1] B) nogil:
    cdef Py_ssize_t i, j, n_rows, n_cols
    cdef double max_sum, row_sum
    
    n_rows = B.shape[0]
    n_cols = B.shape[1]
    
    max_sum = 0.0
    for i in range(n_rows):
        row_sum = 0.0
        for j in range(n_cols):
            row_sum += abs(B[i, j])
        if row_sum > max_sum:
            max_sum = row_sum
    return max_sum

cdef void MM_cdot_pointer_v_scaled(double complex *a0, double complex *v0, double complex *c0, double complex alpha, int n, int m) nogil:
    # matrix multiply 2 matrices A (n x n) and B (n x m)
    cdef char *ori = 'n'
    #cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(ori, ori, &m, &n, &n, &alpha, v0, &m, a0, &n, &beta, c0, &m)
    
cpdef _expm_multiply_simple_core(double complex[:, ::1] A, double complex[:, ::1] B, double complex[:, ::1] F, double complex[:, ::1] G, double t, double mu, int m_star, int s):
    # A (nxn), B (nxm), F (nxm), G (nxm)
    # s is the number of iterations to perform, 
    # Set tolerance to machine precision
    cdef double complex *a0 = &A[0, 0]
    cdef double complex *b0 = &B[0, 0]
    cdef double complex *f0 = &F[0, 0]
    cdef double complex *g0 = &G[0, 0]
    cdef double tol = 2 ** -53 
    cdef int n = B.shape[0]
    cdef int m = B.shape[1]
    cdef int nm = n * m
    cdef double factor
    cdef double complex coeff = 1.0
    cdef double complex eta = 0.0
    copy_pointer(b0, f0, nm)  # Initialize F as B
    eta = exp(t*mu / float(s))  # Compute scaling factor eta
    
    # Perform s iterations
    for i in range(s):
        c1 = norm_inf_complex(B)            # Compute infinity norm of B
        for j in range(m_star):
            factor = float(s*(j+1))
            coeff.real = t / factor
            MM_cdot_pointer_v_scaled(a0, b0, g0, coeff, n, m)  # Update B ( B = coeff * A.dot(B) ) -> switch every iteration between B and G
            # switch pointers of B and G    #Old Approach: C = B; B = G; G = C
            B, G = G, B
            c2 = norm_inf_complex(B)        # Compute norm of updated B
            c_mat_add_pointer(f0, b0, nm)   # Update F ( F= F + B )
            if c1 + c2 <= tol * norm_inf_complex(F):  # Check if tolerance is reached
                break
            c1 = c2
        c_mat_scale(F, eta)  # Scale F
        copy_pointer(f0, b0, nm)  # Copy F to B
    return F  # Return final value of F



#cython: language_level=3
import numpy as np
cimport numpy as npc
from libc.math cimport fabs
from .exp_and_log cimport *
from .indexing cimport *
from .caching cimport *
from .blas_functions cimport *
from .blas_functions_vectors cimport *
from .autobinning import optimal_binning


# Unitary Interpolation
cdef class UI_vector:
    # Initialize variables, to quickly calculate interpolations while minimizing memmory allocation overheads
    cdef double[::1] c_mins, c_maxs, dcs, das
    cdef long[::1] c_bins
    cdef int n_dims, d, d2, n_dims_1, n_d_di_1, n_d_di, m, dm
    cdef double[:, ::1] E
    cdef double complex[::1] Es
    cdef double complex[:,:,::1] Vr, Vl, CL, CH, dU1, dU2
    cdef double complex[:,::1] Ur, Ur1
    cdef double* ei
    cdef double complex *vl
    cdef double complex *vr
    cdef double complex *c
    cdef double complex *ur0
    cdef double complex *ur1
    cdef double complex *du1
    cdef double complex *du2
    cdef double complex *es0
    cdef long[::1] strides_L, strides_R,
    cdef long[:,::1] strides_E, strides_C
    cdef long[::1] location, d_location
    cdef double[::1] abs_alpha_rest, alpha
    cdef long[::1] first_elements_E, first_elements_C, L
    cdef long[::1] d_di
    def __cinit__(self, double complex[:,:,::1] H_s, double[::1] c_min_s, double[::1] c_max_s, long[::1] c_bins, long[::1] which_diffs = np.array([], dtype=long), int m = 1):
        # Construct parameters
        self.n_dims = c_min_s.shape[0]
        self.n_dims_1 = self.n_dims - 1
        self.d = H_s.shape[1]
        self.d2 = self.d * self.d
        if not H_s.shape[0] == self.n_dims + 1:
            print('Requires n+1 Hamiltonians for n dimensional interpolation. Check lenths of Hs, c_mins, c_maxs, c_bins')
            raise ValueError
        self.c_bins = np.empty(self.n_dims, dtype=long)
        for i in range(self.n_dims):
            self.c_bins[i] = c_bins[i]
        self.c_mins, self.c_maxs, self.dcs, self.das = Bin_Parameters(c_min_s, c_max_s, self.c_bins)
        # Single Step Indexing Parameters  --> Add multiple indexes later
        self.location = np.empty(self.n_dims, dtype=long)
        self.abs_alpha_rest = np.empty(self.n_dims, dtype=np.double)
        self.d_location = np.empty(self.n_dims, dtype=long)
        self.alpha = np.empty(self.n_dims, dtype=np.double)
        if which_diffs.shape[0] == 0:
            self.d_di = np.arange(self.n_dims)
        else:
            self.d_di = which_diffs
        self.n_d_di = self.d_di.shape[0]

        # Construct interpolation grid points
        U_grid, cum_prod = Unitary_Grid(H_s, self.c_mins, self.dcs, self.c_bins)
        ## Construct interpolation cache
        self.E, self.Vl, self.Vr, self.CL, self.CH, self.strides_E, self.strides_L, self.strides_R, self.strides_C, self.first_elements_E, self.first_elements_C = Create_Interpolation_Cache( U_grid, cum_prod, self.c_bins)
        self.ei = &self.E[0, 0]
        self.vl = &self.Vl[0, 0, 0]
        self.vr = &self.Vr[0, 0, 0]
        self.c = &self.CH[0, 0, 0]
        self.L = np.empty(self.n_dims, dtype=long)
        self.change_m(m)

    def change_m(self, int m):
        self.m = m
        self.dm = self.d * m
        # also change self.vi's
        self.Ur = np.empty([self.d, self.m], dtype=np.complex128)
        self.ur0 = &self.Ur[0, 0]
        self.Ur1 = np.empty([self.d, self.m], dtype=np.complex128)
        self.ur1 = &self.Ur1[0, 0]
        self.dU1 = np.empty([self.n_d_di, self.d, self.m], dtype=np.complex128)
        self.du1 = &self.dU1[0, 0, 0]
        self.dU2 = np.empty([self.n_d_di, self.d, self.m], dtype=np.complex128)
        self.du2 = &self.dU2[0, 0, 0]
        self.Es = np.empty(self.d, dtype=np.complex128)
        self.es0 = &self.Es[0]

    cdef single_parameters2oddgrid(self, double[::1] c):
        cdef int sum_location = 0
        cdef double alpha_max = 0.0
        cdef double alpha_rest
        cdef Py_ssize_t i
        cdef int max_alpha_ind = 0
        cdef int max_vals
        for i in range(self.n_dims):
            # Transform
            self.alpha[i] = (c[i] - self.c_mins[i]) / self.dcs[i]
            # Round
            self.location[i] = <int> self.alpha[i]     # floor -> 
            alpha_rest = self.alpha[i] - self.location[i]
            if alpha_rest > 0.5:
                self.location[i] += 1
                alpha_rest -= 1
            sum_location += self.location[i]

            self.abs_alpha_rest[i] = fabs(alpha_rest)
            self.d_location[i] = 1 if alpha_rest >= 0 else -1
            if self.alpha[i] < 0 or self.alpha[i] > self.c_bins[i]:
                print('Warning: These parameters lie outside of interpolation grid!')
                break
        if is_even(sum_location):
            max_alpha_ind = c_argmax(self.abs_alpha_rest, max_alpha_ind, alpha_max, self.n_dims)
            if self.location[max_alpha_ind] + self.d_location[max_alpha_ind] > self.c_bins[max_alpha_ind]:
                self.d_location[max_alpha_ind] = -1
            self.location[max_alpha_ind] += self.d_location[max_alpha_ind]
            self.d_location[max_alpha_ind] = - self.d_location[max_alpha_ind]
            self.abs_alpha_rest[max_alpha_ind] = 1-self.abs_alpha_rest[max_alpha_ind]
        for i in range(self.n_dims):
            self.d_location[i] -= 1
            self.d_location[i] >>= 1 # via bitshift operation
            #d_location[i] /= 2 #(d_location[i]-1)/2 ### in two steps with optimized in-place operations
        for i in range(self.n_dims):
            max_vals = self.location[i]+self.d_location[i]
            if max_vals > self.c_bins[i]-1:
                self.d_location[i] = -1

    def set_which_diffs(self, long[::1] which_diffs):
        self.d_di = which_diffs
        self.n_d_di_1 = self.d_di.shape[0] - 1
        self.n_d_di = self.d_di.shape[0]

    def get_single_param(self, double[::1] c): # To verify interpolation grid
        self.single_parameters2oddgrid(c)
        return np.asarray(self.location), np.asarray(self.d_location), np.asarray(self.abs_alpha_rest)

    def get_cache(self): # To verify interpolation grid
        return np.asarray(self.E), np.asarray(self.Vl), np.asarray(self.Vr), np.asarray(self.CL), np.asarray(self.CH), np.asarray(self.strides_E), np.asarray(self.strides_L), np.asarray(self.strides_R), np.asarray(self.strides_C)

    cdef interpolate_single_u(self, double complex *u0, double complex *v0): #u0 => input vectors, v0 => output vectors
        cdef Py_ssize_t i, j
        cdef Py_ssize_t ind
        if self.n_dims == 1:
            ind = self.location[0] + self.d_location[0]
            self.vr = &self.Vr[ind,0,0] # In 1D strides are 1
            self.vl = &self.Vl[ind,0,0]
            self.ei = &self.E[ind,0]
            MM_cdot_pointer_v(self.vr, u0, self.ur0, self.d, self.m)
            v_exp_pointer_v(self.ei, self.ur0, self.abs_alpha_rest[0], self.d, self.m)
            MM_cdot_pointer_v(self.vl, self.ur0, v0, self.d, self.m)
        else:
            # Right side first
            self.L[0] = self.location[0] + self.d_location[0]
            for i in range(1, self.n_dims):
                self.L[i] = self.location[i]
            ind = findex_0(self.L, self.strides_E[0,:], self.n_dims)
            self.vr = &self.Vr[ind, 0, 0]  # In 1D strides are 1
            MM_cdot_pointer_v(self.vr, u0, self.ur0, self.d, self.m)
            self.ei = &self.E[ind, 0]
            v_exp_pointer_v(self.ei, self.ur0, self.abs_alpha_rest[0], self.d, self.m)
            self.L[0] = self.location[0]
            # Center multiplications
            for i in range(self.n_dims_1):
                j = i + 1
                self.L[j] += self.d_location[j]
                ind = findex_0(self.L, self.strides_E[j, :], self.n_dims) + self.first_elements_E[j]
                self.ei = &self.E[ind, 0]
                if self.d_location[i]:  ### Higher
                    self.L[i] += -1
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims) + self.first_elements_C[i]
                    self.c = &self.CL[ind, 0, 0]
                    self.L[i] += 1  # Restore value
                else:                        ### Lower
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims)+self.first_elements_C[i]
                    self.c = &self.CH[ind,0,0]
                self.L[j] = self.location[j]
                MM_cdot_pointer_v(self.c, self.ur0, self.ur1, self.d, self.m)
                v_exp_pointer_v(self.ei, self.ur1, self.abs_alpha_rest[j], self.d, self.m)
                self.ur1, self.ur0 = self.ur0, self.ur1
            # Left side multiplication
            self.L[self.n_dims_1] += self.d_location[self.n_dims_1]
            ind = findex_0(self.L, self.strides_L, self.n_dims)
            self.vl = &self.Vl[ind,0,0]
            MM_cdot_pointer_v(self.vl, self.ur0, v0, self.d, self.m)

    cdef interpolate_single_u_du(self, double complex *u0, double complex *v0, double complex *du0): #u0, du0 => input the matrices for output   ##int[::1] d_di, -> define earlier
        # d_di contains the indexes of the derivatives that we want to calculate
        cdef Py_ssize_t i, j, ind, curr_d_di
        cdef int curr_d_ind = 0
        cdef double complex *du1 = self.du1 #.copy()
        cdef double complex *du2 = self.du2 #.copy()
        cdef double complex *ur0 = self.ur0
        cdef double complex *ur1 = self.ur1
        cdef double complex *es0 = self.es0
        if self.n_dims == 1:
            ind = self.location[0] + self.d_location[0]
            self.vr = &self.Vr[ind, 0, 0] # In 1D strides are 1
            self.vl = &self.Vl[ind, 0, 0]
            self.ei = &self.E[ind, 0]
            MM_cdot_pointer_v(self.vr, u0, ur0,  self.d, self.m)
            if self.d_di[curr_d_ind] == 0:
                copy_pointer(ur0, du1, self.dm)
                v_exp_v_pointer_v( (self.d_location[0]*2+1)*self.das[0], self.ei, du1, self.abs_alpha_rest[0], self.d, self.m) # -i*E*amp*exp(-i*E*t)
                MM_cdot_pointer_v(self.vl, self.du1, du0, self.d, self.m)
            v_exp_pointer_v(self.ei, ur0, self.abs_alpha_rest[0], self.d, self.m)
            MM_cdot_pointer_v(self.vl, ur0, v0, self.d, self.m)
        else:
            # Track the number of performed derivatives via curr_d_ind, the indexes to perform with self.d_di and the next derivative to perform with self.d_di[curr_d_ind] == curr_d_di
            # Right side first
            self.L[0] = self.location[0] + self.d_location[0]
            for i in range(1, self.n_dims):
                self.L[i] = self.location[i]
            ind = findex_0(self.L, self.strides_E[0,:], self.n_dims)
            self.vr = &self.Vr[ind, 0, 0]  # In 1D strides are 1
            self.ei = &self.E[ind, 0]

            MM_cdot_pointer_v(self.vr, u0, ur0, self.d, self.m) # now ur0 and du1
            curr_d_di = self.d_di[curr_d_ind]
            if curr_d_di == 0:
                copy_pointer(ur0, du1, self.dm)
                v_exp_v_pointer_v( (self.d_location[0]*2+1)*self.das[0], self.ei, du1, self.abs_alpha_rest[0], self.d, self.m) # -i*E*amp*exp(-i*E*t)
                curr_d_ind += 1
                curr_d_di = self.d_di[curr_d_ind]
            v_exp_pointer_v(self.ei, ur0, self.abs_alpha_rest[0], self.d, self.m) # still ur0 and du1

            self.L[0] = self.location[0]
            # Center multiplications
            for i in range(self.n_dims_1):
                j = i + 1
                self.L[j] += self.d_location[j]
                ind = findex_0(self.L, self.strides_E[j, :], self.n_dims) + self.first_elements_E[j]
                self.ei = &self.E[ind, 0]
                if self.d_location[i]:  ### Higher
                    self.L[i] += -1
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims) + self.first_elements_C[i]
                    self.c = &self.CL[ind, 0, 0]
                    self.L[i] += 1  # Restore value
                else:                        ### Lower
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims)+self.first_elements_C[i]
                    self.c = &self.CH[ind,0,0]
                self.L[j] = self.location[j]
                # Batch matrix multiply
                MM_cdot_pointer_batch_v(self.c, ur0, ur1, du1, du2, self.d, self.m, self.dm, curr_d_ind) # now ur1 and du2
                # Perform the row wise complex scaling, with potential derivative
                if curr_d_di == j: # Add derivative here
                    v_exp_v_and_v_exp_pointer_v_batch_increment((self.d_location[curr_d_di]*2+1)*self.das[curr_d_di], self.ei, es0, ur1, du2, self.abs_alpha_rest[j], self.d, self.m, self.dm, curr_d_ind)
                    curr_d_ind += 1
                    curr_d_di = self.d_di[curr_d_ind]
                else:
                    v_exp_v_and_v_exp_pointer_v_batch(self.ei, self.es0, ur1, du2, self.abs_alpha_rest[j], self.d, self.m, curr_d_ind)
                ur1, ur0 = ur0, ur1 # now flip them
                du2, du1 = du1, du2
            # Left side multiplication
            self.L[self.n_dims_1] = self.location[self.n_dims_1] + self.d_location[self.n_dims_1]
            ind = findex_0(self.L, self.strides_L, self.n_dims)
            self.vl = &self.Vl[ind,0,0]
            MM_cdot_pointer_batch_v(self.vl, ur0, v0, du1, du0, self.d, self.m, self.dm, curr_d_ind) # now ur1 and du2
            

    cdef expmH_pointer(self, double[::1] c, double complex *u0, double complex *v0):
        self.single_parameters2oddgrid(c)
        self.interpolate_single_u(u0, v0)
    def expmH(self, double[::1] c, double complex[:,::1] V_in, double complex[:,::1] V_out):
        if not c.shape[0] == self.n_dims:
            raise ValueError('c.shape[0] needs to be equal to H_s[0].shape[0]-1.')
        # check V_in size
        if not V_in.shape[0] == self.d:
            raise ValueError('V_in.shape[0] needs to be equal to H_s[0].shape[0].')
        if not V_in.shape[1] == self.m:
            # change m
            self.change_m(V_in.shape[1])
        # check V_out size
        if not V_out.shape[0] == self.d:
            raise ValueError('V_out.shape[0] needs to be equal to H_s[0].shape[0].')
        if not V_out.shape[1] == self.m:
            raise ValueError('V_out.shape[1] needs to be equal to V_in.shape[1].')
        cdef double complex *u0 = &V_in[0, 0]
        cdef double complex *v0 = &V_out[0, 0]
        self.expmH_pointer(c, u0, v0)

    cdef dexpmH_pointer(self, double[::1] c, double complex *u0, double complex *v0, double complex *du0):  #int[::1] d_di,
        # d_di contains the indexes of the derivatives that we want to calculate (needs to be in ascending order with a negative value at the end)
        self.single_parameters2oddgrid(c)
        self.interpolate_single_u_du(u0, v0, du0)

    def dexpmH(self, double[::1] c, double complex[:,::1] V_in, double complex[:,::1] V_out, double complex[:,:,::1] dU):  #int[::1] d_di,
        # d_di contains the indexes of the derivatives that we want to calculate (needs to be in ascending order with a negative value at the end)
        cdef double complex *u0 = &V_in[0,0]
        cdef double complex *v0 = &V_out[0,0]
        cdef double complex *du0 = &dU[0,0,0]
        if not c.shape[0] == self.n_dims:
            raise ValueError('The coefficient c must be of size [interpolation_dimensions].')
        if not self.n_d_di == dU.shape[0]:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = dU.shape[0].')
        if not V_in.shape[0] == self.d:
            raise ValueError('The input vector must be of size [d].')
        if not V_out.shape[0] == self.d:
            raise ValueError('The output vector must be of size [d].')
        if not V_in.shape[1] == V_out.shape[1] == self.m:
            raise ValueError('The input and output vectors must have the same number of columns, and equal to m.')
        if not dU.shape[1] == self.d:
            raise ValueError('The input derivative vector must be of size [d].')
        if not dU.shape[2] == self.m:
            raise ValueError('The input derivative vector must have the same number of columns as the input and output vectors, and equal to m.')
        if not self.n_d_di == dU.shape[0]:
            raise ValueError('The input derivative vector must have the same number of rows as the number of derivatives to calculate.')
        self.dexpmH_pointer(c, u0, v0, du0)


    cdef expmH_pulse_pointer(self, double[:,::1] cs, double complex *u0, double complex *v0):
        cdef Py_ssize_t i = 0
        cdef int steps = cs.shape[0]
        cdef int steps_2 = steps >> 1  # floor (steps/2)
        cdef int is_odd_steps = steps % 2
        if not cs.shape[1] == self.n_dims:
            raise ValueError('The coefficient matrix must be of size [n_time_steps, interpolation_dimensions].')
        for i in range(0, steps_2, 2):
            self.single_parameters2oddgrid(cs[i,:])
            self.interpolate_single_u(u0, v0)
            self.single_parameters2oddgrid(cs[i+1,:])
            self.interpolate_single_u(v0, u0)
        if is_odd_steps:
            self.single_parameters2oddgrid(cs[steps_2,:])
            self.interpolate_single_u(u0, v0)
        else:  # copy to v0 
            copy_pointer(u0, v0, self.dm)

    def expmH_pulse(self, double[:,::1] cs, double complex[:,::1] V_in, double complex[:,::1] V_out):
        cdef double complex *u0 = &V_in[0, 0]
        cdef double complex *v0 = &V_out[0, 0]
        self.expmH_pulse_pointer(cs, u0, v0)


    def expmH_pulse_no_multiply(self, double[:,::1] cs, double complex[:,::1] V_in, double complex[:,:,::1] V_out):
        cdef double complex *u0 = &V_in[0,0]
        cdef double complex *v0 = &V_out[0,0,0]
        cdef n = cs.shape[0]
        for i in range(n):
            self.expmH_pointer(cs[i,:], u0, v0)
            v0 += self.dm
    """
    def grape(self, double[:,::1] cs, double complex[:,::1] U_target, int[::1] target_indexes, double complex[:,::1] U, double complex[:,:,::1] dU, double[:,::1] dI_dj):
        # Calculate fidelity for a pulse and the differentials of the fidelity at every timestep using the grape trick
        cdef Py_ssize_t i, j
        cdef int steps = cs.shape[0]

        cdef double complex *new_p0 = self.ur0
        cdef double complex *new_q0 = self.ur1
        cdef double complex *p0 = self.ur4
        cdef double complex *q0 = self.ur5
        cdef double complex *u0 = &U[0, 0]
        cdef double complex *u_tar = &U_target[0, 0]
        cdef double complex *curr_u = self.ur3
        cdef double complex *du = &dU[0,0,0]

        cdef double complex trM, trdM, two_nni
        cdef double n = <double> target_indexes.shape[0]
        cdef double nni = 1.0 / (n * (n + 1))
        cdef double one_minus = 1.0 - n * nni

        if not dI_dj.shape[0] == steps:
            raise ValueError('Inputs must fulfill: cs.shape[0] = dI_dj.shape[0].')
        if not dI_dj.shape[1] == self.n_d_di:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = dI_dj.shape[1].')
        if not cs.shape[1] == self.n_d_di:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = cs.shape[1].')
        # Fidelity constants

        self.expmH_pulse_pointer(cs, u0)
        #Dag_fast(U_target, self.Ur3) # -> Ur3 is identical to Q  -> replaced by new approach
        DagM_M_cdot_pointer(u_tar, u0, q0, self.d)
        trM = target_indexes_trace_pointer(q0, self.d, target_indexes)
        two_nni = - 2 / (n * (n + 1)) * conj(trM)
        cdef double I0 = one_minus - nni * abs_2(trM) # Sub Fidelity

        # Do GRAPE trick
        self.dexpmH_pointer(cs[0,:], curr_u, du) # Calculate Unitary and its derivatives
        M_DagM_cdot_pointer(q0, curr_u, new_q0, self.d)  # Reduce q0
        copy_pointer(new_q0, q0, self.d2)
        for j in range(self.n_d_di):
            #MM_cdot_pointer(curr_du, p0, new_p0, self.d)  # p0 is not defined yet
            trdM = tr_dot_pointer_target_indexes(q0, du, self.d, target_indexes)
            dI_dj[0, j] = ( two_nni * trdM ).real
            du += self.d2
        copy_pointer(curr_u, p0, self.d2)  # Define p0 from curr_u
        for i in range(1, steps):
            du -= self.d2 * self.n_d_di
            self.dexpmH_pointer(cs[i, :], curr_u, du)  # Calculate Unitary and its derivatives
            M_DagM_cdot_pointer(q0, curr_u, new_q0, self.d)  # Remove current unitary (curr_u) from Q
            for j in range(self.n_d_di):
                MM_cdot_pointer(du, p0, new_p0, self.d)
                trdM = tr_dot_pointer_target_indexes(new_q0, new_p0, self.d, target_indexes)
                dI_dj[i, j] = ( two_nni * trdM ).real
                du += self.d2
            MM_cdot_pointer(curr_u, p0, new_p0, self.d)  # Add current unitary (curr_u) to P
            #copy_pointer(new_p0, p0, self.d2)
            #copy_pointer(new_q0, q0, self.d2)
            # flip pointers
            self.ur0, self.ur4 = self.ur4, self.ur0
            self.ur1, self.ur5 = self.ur5, self.ur1
            new_p0, p0 = p0, new_p0
            new_q0, q0 = q0, new_q0
        return I0
    """


def UI_vector_auto(H_s, c_min_s, c_max_s, I_tar=1e-10, which_diffs = np.array([], dtype=np.compat.long), m = 1):
    opt_bins  = optimal_binning(H_s, c_mins=c_min_s, c_maxs=c_max_s, I_tar=I_tar)
    return UI_vector(H_s, c_min_s, c_max_s, opt_bins, which_diffs, m = 1)



#cython: language_level=3
from cython.operator import dereference
from libc.math cimport cos, sin, log
import numpy as np
#cimport numpy as np
from scipy.linalg.cython_blas cimport zgemm, zscal, zcopy
from scipy.sparse.linalg._expm_multiply import LazyOperatorNormInfo, _fragment_3_1

cdef void MM_cdot_pointer_v_scaled(double complex *a0, double complex *v0, double complex *c0, double complex alpha, int n, int m) nogil:
    # matrix multiply 2 matrices A (n x n) and B (n x m)
    cdef char *ori = 'n'
    #cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(ori, ori, &m, &n, &n, &alpha, v0, &m, a0, &n, &beta, c0, &m)

cdef void MM_cdot_pointer_v(double complex *a0, double complex *v0, double complex *c0, int n, int m) nogil:
    # matrix multiply 2 matrices A (n x n) and B (n x m)
    cdef char *ori = 'n'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(ori, ori, &m, &n, &n, &alpha, v0, &m, a0, &n, &beta, c0, &m)

cpdef void MM_cdot_v(double complex[:,::1] A, double complex[:,::1] v, double complex[:,::1] C): # C = A @ B
    # matrix multiply 2 matrices A (n x n) and B (n x m) -> C = A @ B
    cdef double complex *a0=&A[0,0]
    cdef double complex *v0=&v[0,0]
    cdef double complex *c0=&C[0,0]
    cdef int n = A.shape[0] # = v.shape[0]
    cdef int m = v.shape[1]
    #Mv_cdot_pointer(a0, v0, c0, n, m)
    cdef char *ori = 'n'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(ori, ori, &m, &n, &n, &alpha, v0, &m, a0, &n, &beta, c0, &m)

cdef void DagM_M_cdot_pointer_v(double complex *a0, double complex *v0, double complex *c0, int n, int m) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'n'
    cdef char *orientB = 'c'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(orientA, orientB, &m, &n, &n, &alpha, v0, &m, a0, &n, &beta, c0, &m)

cpdef void DagM_M_cdot_v(double complex[:,::1] A, double complex[:,::1] v, double complex[:,::1] C) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef double complex *a0=&A[0,0]
    cdef double complex *v0=&v[0,0]
    cdef double complex *c0=&C[0,0]
    cdef int n = A.shape[0]
    cdef int m = v.shape[1]
    DagM_M_cdot_pointer_v(a0, v0, c0, n, m)


cdef void v_exp_pointer_v(double *e0, double complex *v0, double t, int n, int m) nogil: # -i*E*amp*exp(-i*E*t)@v
    cdef double complex c
    cdef double ei
    cdef int incz = 1
    cdef int i
    ei = dereference(e0) * t
    c.real = cos(ei)
    c.imag = -sin(ei)
    zscal(&m, &c, v0, &incz)
    for i in range(1, n):
        v0 += m
        e0 += 1
        ei = dereference(e0) * t
        c.real = cos(ei)
        c.imag = -sin(ei)
        zscal(&m, &c, v0, &incz)

cdef void v_exp_v_pointer_v(double amp, double *e0, double complex *v0, double t, int n, int m) nogil: 
    cdef double complex c
    cdef double ei, f
    cdef int incz = 1
    cdef int i

    ei = dereference(e0) * t
    f = -dereference(e0) * amp
    c.real = f * sin(ei)
    c.imag = f * cos(ei)
    zscal(&m, &c, v0, &incz)
    for i in range(1, n):
        v0 += m
        e0 += 1
        ei = dereference(e0) * t
        f = -dereference(e0) * amp
        c.real = f * sin(ei)
        c.imag = f * cos(ei)
        zscal(&m, &c, v0, &incz)



##### Batch computations ############################################
cdef void MM_cdot_pointer_batch_v(double complex *a0, double complex *vi, double complex *vo, double complex *di, double complex *do, int n, int m, int nm, int batch_ind) nogil:
    # calculate all matrix products in a batch jumping by nm for batch_ind 
    cdef char *ori = 'n'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    # copy pointers
    cdef double complex *di2 = di 
    cdef double complex *do2 = do
    zgemm(ori, ori, &m, &n, &n, &alpha, vi, &m, a0, &n, &beta, vo, &m)
    for i in range(batch_ind):
        zgemm(ori, ori, &m, &n, &n, &alpha, di2, &m, a0, &n, &beta, do2, &m)
        di2 += nm
        do2 += nm

cdef void MM_cdot_batch_v(double complex[:,::1] A, double complex[:,::1] V_in, double complex[:,::1] V_out, double complex[:,:,::1] dV_in, double complex[:,:,::1] dV_out, int n, int m, int nm, int batch_ind) nogil:
    # calculate all matrix products in a batch jumping by nm for batch_ind 
    cdef double complex *a0=&A[0,0]
    cdef double complex *vi=&V_in[0,0]
    cdef double complex *vo=&V_out[0,0]
    cdef double complex *di=&dV_in[0,0,0]
    cdef double complex *do=&dV_out[0,0,0]
    MM_cdot_pointer_batch_v(a0, vi, vo, di, do, n, m, nm, batch_ind)
    
cdef void v_exp_v_and_v_exp_pointer_v_batch(double *e0, double complex *e0s, double complex *v0, double complex *dv0, double t, int n, int m, int batch_ind) nogil: 
    # use matrix exponential for every page in dv0
    # if batch_ind_increment=1 add a last element, copy the first element to the last element
    cdef double complex c, g
    cdef double ei
    cdef double complex f
    cdef int incz = 1
    cdef Py_ssize_t i, j
    # first calculate the array e0s once
    for i in range(n):
        #multiply the n'th element of e0 with t
        ei = t * dereference(e0)
        c.real = cos(ei)
        c.imag = -sin(ei)
        e0s[i] = c
        e0 += 1
    # now calculate the row wise exponential first of v0 and then of the pages of dv0
    for i in range(n):
        zscal(&m, e0s, v0, &incz)
        v0 += m
        e0s += 1
    # now calculate the row wise exponential of dv0
    for i in range(batch_ind): # page wise
        e0s -= n
        for j in range(n):
            zscal(&m, e0s, dv0, &incz)
            dv0 += m
            e0s += 1

cpdef void v_exp_v_and_v_exp_v_batch(double[::1] E, double complex[::1] Es, double complex[:,::1] V, double complex[:,:,::1] dV, double t, int batch_ind):
    cdef double *e0 = &E[0]
    cdef double complex *e0s = &Es[0]
    cdef double complex *v0 = &V[0,0]
    cdef double complex *dv0 = &dV[0,0,0]
    cdef int m = V.shape[1]
    cdef int n = V.shape[0]
    v_exp_v_and_v_exp_pointer_v_batch(e0, e0s, v0, dv0, t, n, m, batch_ind)

cdef void v_exp_v_and_v_exp_pointer_v_batch_increment(double amp, double *e0, double complex *e0s, double complex *v0, double complex *dv0, double t, int n, int m, int mn, int batch_ind) nogil: 
    # use matrix exponential for every page in dv0
    # if batch_ind_increment=1 add a last element, copy the first element to the last element
    cdef double complex c, g
    cdef double ei
    cdef double complex f
    cdef int incz = 1
    cdef Py_ssize_t i, j
    cdef int k = mn * batch_ind

    # first calculate the array e0s once
    for i in range(n):
        #multiply the n'th element of e0 with t
        ei = t * dereference(e0)
        c.real = cos(ei)
        c.imag = -sin(ei)
        e0s[i] = c
        e0 += 1
    # now calculate the row wise exponential first of v0 and then of the pages of dv0
    for i in range(n):
        zscal(&m, e0s, v0, &incz)
        v0 += m
        e0s += 1
    # precopy -> then rescale later
    v0 -= mn
    dv0 += k
    zcopy(&mn, v0, &incz, dv0, &incz)
    dv0 -= k
    # now calculate the row wise exponential of dv0
    for i in range(batch_ind): # page wise
        e0s -= n
        for j in range(n):
            zscal(&m, e0s, dv0, &incz)
            dv0 += m
            e0s += 1
    # scale the precopied elements
    e0 -= n
    f.real = 0.0
    for i in range(n): # copy and scale by f
        f.imag = - amp * dereference(e0)
        # complex multiplication
        zscal(&m, &f, dv0, &incz)
        dv0 += m
        e0 += 1

cpdef void v_exp_v_and_v_exp_v_batch_increment(double amp, double[::1] E, double complex[::1] Es, double complex[:,::1] V, double complex[:,:,::1] dV, double t, int batch_ind):
    cdef double *e0 = &E[0]
    cdef double complex *e0s = &Es[0]
    cdef double complex *v0 = &V[0,0]
    cdef double complex *dv0 = &dV[0,0,0]
    cdef int m = V.shape[1]
    cdef int n = V.shape[0]
    cdef int mn = m * n
    v_exp_v_and_v_exp_pointer_v_batch_increment(amp, e0, e0s, v0, dv0, t, n, m, mn, batch_ind)

#### Not really blas functions, but added here nonetheless
cdef double norm_inf_complex( double complex[:, ::1] B, int d, int m) nogil:
    cdef Py_ssize_t i, j, n_rows, n_cols
    cdef double row_sum
    cdef double max_sum = 0.0
    for i in range(d):
        row_sum = 0.0
        for j in range(m):
            row_sum += abs(B[i, j])
        if row_sum >  max_sum:
                max_sum = row_sum
    return max_sum

cpdef expm_multiply_prepare(double complex[:,::1] A0, int m, double tol):
    """
    Prepare the computation of the action of the matrix exponential of A on B.

    Parameters
    ----------
    A : transposable linear operator
        The operator whose exponential is of interest.
    m : number of wavevectors to use (B will be n x m)
    t : float (default=1.0)
    References
    ----------
    .. [1] Awad H. Al-Mohy and Nicholas J. Higham (2011)
           "Computing the Action of the Matrix Exponential,
           with an Application to Exponential Integrators."
           SIAM Journal on Scientific Computing,
           33 (2). pp. 488-511. ISSN 1064-8275
           http://eprints.ma.man.ac.uk/1591/

    .. [2] Nicholas J. Higham and Awad H. Al-Mohy (2010)
           "Computing Matrix Functions."
           Acta Numerica,
           19. 159-208. ISSN 0962-4929
           http://eprints.ma.man.ac.uk/1451/
    """
    # Check the inputs.
    A = np.array(A0)
    #print(A.shape)
    if A.shape[0] != A.shape[1]:
        raise ValueError('expected A to be like a square matrix')
    ident = np.eye(A.shape[0], A.shape[1], dtype=np.complex128)
    n = A.shape[0]
    traceA =  np.array(A).trace()
    mu = traceA / float(n)
    A = A - mu * ident
    A_1_norm = np.linalg.norm(A, 1)  
    if A_1_norm == 0:
        m_star, s = 0, 1
    else:
        ell = 2
        norm_info = LazyOperatorNormInfo(A, A_1_norm=A_1_norm)
        m_star, s = _fragment_3_1(norm_info, m, tol)
    return m_star, s, np.real(mu)


# -*- coding: utf-8 -*-
#from unipolator.autobinning import *
from unipolator.hamiltonian_system import *

from unipolator.unitary_interpolation import *
from unipolator.symmetric_unitary_interpolation import *
from unipolator.trotter_system import *
from unipolator.sym_trotter_system import *

from unipolator.unitary_interpolation_vector import *
from unipolator.symmetric_unitary_interpolation_vector import *

from unipolator.trotter_system_vector import *
from unipolator.sym_trotter_system_vector import *
from unipolator.hamiltonian_system_vector import *



#cython: language_level=3
import numpy as np
cimport numpy as npc
from libc.math cimport exp
from .blas_functions cimport *
from .blas_functions_vectors cimport *
from .exp_and_log cimport copy_pointer 
from scipy.linalg.cython_blas cimport zgemm, zaxpy, zscal

# Unitary Interpolation
cdef class Hamiltonian_System_vector:
    # Initialize variables, to quickly calculate interpolations while minimizing memmory allocation overheads
    cdef int n_dims, n_dims_1, d, d2, m, dm
    cdef double complex[:,:,::1] H
    cdef double complex[:,::1] V, V_inter, V_in
    cdef double complex *h0
    cdef double complex *v0
    cdef double complex *v_inter
    cdef double complex *v_in
    cdef double complex *curr_h0
    cdef double tol
    cdef double[::1] max_amp
    cdef int m_star, s
    cdef double mu
    
    def __cinit__(self, double complex[:,:,::1] H_s, int m=1, double tol=2**-53, double[::1] max_amp=np.array([-1.0])):
        # Construct exponential using scaling and squaring
        # H_s is a 3D array of shape (n_dims, d, d) where n_dims is the number of dimensions of the Hamiltonian and d is the dimension of the Hilbert space
        # m is the number of wavevectors to compute at once (can be changed using change_m)
        # tol is the tolerance of the approximation (can be changed using change_tolerance)
        # max_amp are the maximum coefficient of the Hamiltonian parts (1.0 by default for each dimension)

        # Construct parameters
        self.n_dims = H_s.shape[0]
        self.n_dims_1 = self.n_dims - 1
        self.d = H_s.shape[1]
        self.d2 = self.d * self.d
        self.m = m
        if max_amp[0] == -1:
            self.max_amp = np.ones(self.n_dims)
        else:
            self.max_amp = max_amp

        self.H = np.empty([self.n_dims, self.d, self.d], dtype=np.complex128)
        self.h0 = &self.H[0,0,0]
        # change elements in H to -i * H
        for i in range(self.n_dims):
            for j in range(self.d):
                for k in range(self.d):
                    self.H[i, j, k] = -1j * H_s[i, j, k]
        
        self.V = np.empty([self.d, self.d], dtype=np.complex128)
        self.v0 = &self.V[0, 0]

        self.curr_h0 = &self.H[0, 0, 0]
        self.change_m(m)
        self.change_tolerance(tol)


    def change_m(self, int m):
        self.m = m
        self.dm = self.d * m
        self.V_in = np.empty([self.d, self.dm], dtype=np.complex128)
        self.v_in = &self.V_in[0, 0]
        self.V_inter = np.empty([self.d, self.dm], dtype=np.complex128)
        self.v_inter = &self.V_inter[0, 0]

    def change_tolerance(self, double tol):
        self.tol = tol
        self.weighted_hamiltonian(self.max_amp)
        self.m_star, self.s, self.mu = expm_multiply_prepare(self.V, self.m, self.tol)
    
    def get_parameters(self):
        return self.m_star, self.s, self.mu

    def set_m_star(self, int m_star):
        self.m_star = m_star
    
    def set_s(self, int s):
        self.s = s
    
    def set_mu(self, double mu):
        self.mu = mu


    cdef weighted_hamiltonian(self, double[::1] c):
        self.curr_h0 = self.h0
        copy_pointer(self.curr_h0, self.v0, self.d2)
        for i in range(self.n_dims_1):
            self.curr_h0 += self.d2
            d_mat_add_pointer(self.v0, self.curr_h0, c[i], self.d2)


    cdef void expm_multiply_cython(self, double complex[:, ::1] V_out) nogil:
        # A (nxn), B (nxm), F (nxm), G (nxm)
        # s is the number of iterations to perform, 
        # Set tolerance to machine precision
        # V0 -> hamiltonian (weighted)
        cdef double complex *vout = &V_out[0, 0] # V_out
        cdef int temp_int, i, j
        cdef int incz = 1
        cdef double factor, c1, c2, max_amp
        cdef double complex coeff = 1.0
        cdef double complex eta = exp(self.mu / float(self.s))  # Compute scaling factor eta

        copy_pointer(self.v_in, vout, self.dm)  # Initialize F as B
        # Perform s iterations
        for i in range(self.s):
            c1 = norm_inf_complex(self.V_in, self.d, self.m)            # Compute infinity norm of V_in
            for j in range(self.m_star):
                temp_int = self.s*(j+1)
                factor = <double> float(temp_int)
                coeff = 1 / factor   #.real
                # preform matrix multiplication G = A @ B 
                MM_cdot_pointer_v_scaled(self.v0, self.v_in, self.v_inter, coeff, self.d, self.m)  # Update B ( B = coeff * A.dot(B) ) -> switch every iteration between B and G
                # switch pointers of B and G    #Old Approach: C = B; B = G; G = C
                copy_pointer(self.v_inter, self.v_in, self.dm)
                c2 = norm_inf_complex(self.V_in, self.d, self.m)        # Compute norm of updated B
                c_mat_add_pointer(vout, self.v_in, self.dm)   # Update F ( F= F + B )
                max_amp = norm_inf_complex(V_out, self.d, self.m)
                if c1 + c2 <= self.tol * max_amp:  # Check if tolerance is reached
                    break
                c1 = c2
            zscal(&self.dm, &eta, vout, &incz)  # Scale v_out by eta
            copy_pointer(vout, self.v_in, self.dm)  # Copy v_out to v_in


    def expmH(self, double[::1] c, double complex[:,::1] V_in, double complex[:,::1] V_out):
        # Construct Hamiltonian
        if not c.shape[0] == self.n_dims_1:
            raise ValueError('c.shape[0] needs to be equal to H_s[0].shape[0]-1.')
        if not V_in.shape[0] == V_out.shape[0] == self.d:
            raise ValueError('V_in.shape[0] and V_out.shape[0] needs to be equal to H_s[0].shape[1].')
        if not V_in.shape[1] == V_out.shape[1] == self.m:
            raise ValueError('V_in.shape[1] and V_out.shape[1] needs to be equal to m.')
        cdef double complex *vin = &V_in[0, 0] # V_in
        copy_pointer(vin, self.v_in, self.dm)  
        self.weighted_hamiltonian(c)
        self.expm_multiply_cython(V_out)

    def expmH_pulse_no_multiply(self, double[:,::1] cs, double complex[:,::1] V_in, double complex[:,:,::1] V_out):
        # repeatedly calculate expmH for different c and store them in different V_out
        cdef int n = cs.shape[0]
        cdef double complex *vin = &V_in[0, 0] # V_in
        copy_pointer(vin, self.v_in, self.dm)
        for i in range(n):  
            self.weighted_hamiltonian(cs[i,:])
            self.expm_multiply_cython(V_out[i,:,:])

